/******************************************************************************
  (c) Copyright 2002-2008, 1060 Research Ltd                                   

  This Software is licensed to You, the licensee, for use under the terms of   
  the 1060 Public License v1.0. Please read and agree to the 1060 Public       
  License v1.0 [www.1060research.com/license] before using or redistributing   
  this software.                                                               

  In summary the 1060 Public license has the following conditions.             
  A. You may use the Software free of charge provided you agree to the terms   
  laid out in the 1060 Public License v1.0                                     
  B. You are only permitted to use the Software with components or applications
  that provide you with OSI Certified Open Source Code [www.opensource.org], or
  for which licensing has been approved by 1060 Research Limited.              
  You may write your own software for execution by this Software provided any  
  distribution of your software with this Software complies with terms set out 
  in section 2 of the 1060 Public License v1.0                                 
  C. You may redistribute the Software provided you comply with the terms of   
  the 1060 Public License v1.0 and that no warranty is implied or given.       
  D. If you find you are unable to comply with this license you may seek to    
  obtain an alternative license from 1060 Research Limited by contacting       
  license@1060research.com or by visiting www.1060research.com                 

  NO WARRANTY:  THIS SOFTWARE IS NOT COVERED BY ANY WARRANTY. SEE 1060 PUBLIC  
  LICENSE V1.0 FOR DETAILS                                                     

  THIS COPYRIGHT NOTICE IS *NOT* THE 1060 PUBLIC LICENSE v1.0. PLEASE READ     
  THE DISTRIBUTED 1060_Public_License.txt OR www.1060research.com/license      
 *****************************************************************************/
package org.netkernel.container.impl;

import java.text.SimpleDateFormat;
import java.util.*;

import org.netkernel.container.IKernel;
import org.netkernel.container.IKernelListener;
import org.netkernel.container.ILogger;
import org.netkernel.container.IRepresentationCache;
import org.netkernel.container.IResolutionCache;
import org.netkernel.container.ISpaceListener;
import org.netkernel.container.ITimer;
import org.netkernel.container.config.IConfiguration;
import org.netkernel.container.config.IConfigurationListener;
import org.netkernel.request.IRequest;
import org.netkernel.request.IResponse;
import org.netkernel.scheduler.Scheduler;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.ISpaceWithIdentity;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.IVersion;
import org.netkernel.urii.impl.NetKernelException;
import org.netkernel.util.MultiMap;
import org.netkernel.util.Utils;


/** Implementation of IKernel
 * @author  tab
 */
public class Kernel implements IKernel
{
	private final MultiMap mSpaces = new MultiMap(128,2);
	private final List mSpaceListeners = new ArrayList(128);

	private IConfiguration mConfiguration;
	private final List mConfigurationListeners = new ArrayList(128);

	private Scheduler mScheduler;
	private ILogger mLogger;
	private ThreadGroup mRootThreadGroup;
	private boolean mInDebug;
	private IKernelListener mMonitor = new NullKernelListener();
	private IRepresentationCache mRepresentationCache=new NullRepresentationCache();
	private IResolutionCache mResolutionCache=new NullResolutionCache();
	private RequestBlocker mRequestBlocker;
	private boolean mAreSpaceListenersDirty;
	private ITimer mTimer;
	
	/** Creates a new instance of Container */
	public Kernel()
	{	mRootThreadGroup = Thread.currentThread().getThreadGroup();
		mScheduler = new Scheduler(this);
		mRequestBlocker = new RequestBlocker();
	}
	
	public void setLogger(ILogger aLogger)
	{	if (mLogger==null)
		{	String product=getConfiguration().getString("netkernel.instance.product", "NetKernel");
			String major = getConfiguration().getString("netkernel.instance.version.major", "?");
			String minor =getConfiguration().getString("netkernel.instance.version.minor", "?");
			aLogger.log(ILogger.INFO, this, "KERNEL_START",product,major,minor, (new SimpleDateFormat("yyyy")).format(new Date()));
		}
		mLogger = aLogger;
	}
	
	public void setMonitor(IKernelListener aMonitor)
	{	if (aMonitor!=null)
		{	mMonitor=aMonitor;
		}
		else
		{	mMonitor = new NullKernelListener();
		}
	}
	
	public void setConfiguration(IConfiguration aConfiguration)
	{	if (mConfiguration!=null)
		{	getLogger().log(ILogger.INFO,this,"MSG_CONFIGURATION_CHANGED");
		}
		mConfiguration=aConfiguration;
		notifyConfigurationListeners();
		mInDebug=mConfiguration.getBoolean("netkernel.debug",false);
		NetKernelException.setTraceDepth(mConfiguration.getInt("netkernel.exception.javastack",4));
		mTimer=TimerImpl.getInstance();
	}
	
	public void setTimer(ITimer aTimer)
	{	mTimer=aTimer;
	}
	
	private int mState=S_PAUSED;
	
	public int getState()
	{	return mState;
	}
	
	public void setState(int aState)
	{
		long timeout = getConfiguration().getLong("netkernel.scheduler.thread.timeout",1000);
		switch(aState)
		{
			case S_READY:
				if (mState==S_PAUSED)
				{	notifySpaceListeners();
					getLogger().log(ILogger.INFO,this,"KERNEL_READY");
					mRepresentationCache.clear();
					mResolutionCache.clear();
					mScheduler.start(mRootThreadGroup);
					mState=S_READY;
				}
				else throw new IllegalArgumentException();
				break;
				
			case S_OPEN:
				if (mState==S_READY)
				{	if (!waitForPending(timeout))
					{	getLogger().log(ILogger.INFO,this,"KERNEL_WARNING_STILL_ACTIVE");
					}					
					getLogger().log(ILogger.INFO,this,"KERNEL_RESUMED");
					mRequestBlocker.clear();
					mState=S_OPEN;
				}
				else throw new IllegalArgumentException();
				break;
				
			case S_CLOSED:
				if (mState==S_OPEN)
				{	getLogger().log(ILogger.INFO,this,"KERNEL_PAUSING");
					mRequestBlocker.block();
					if (!waitForPending(timeout))
					{	getLogger().log(ILogger.WARNING,this,"KERNEL_STOPPED_FORCE");
						hardStop(timeout);
						mScheduler.start(mRootThreadGroup);
					}
					mState=S_CLOSED;
				}
				else throw new IllegalArgumentException();
				break;
			
			case S_PAUSED:
				if (mState==S_CLOSED)
				{	if (!waitForPending(timeout))
					{	getLogger().log(ILogger.WARNING,this,"KERNEL_STOPPED_FORCE");
					}
					hardStop(timeout);
					getLogger().log(ILogger.INFO,this,"KERNEL_STOPPED");
					mState=S_PAUSED;
				}
				else throw new IllegalArgumentException();
				break;		
			case S_FINAL:
				if (mState==S_PAUSED)
				{	getLogger().log(ILogger.INFO,this,"KERNEL_FINAL");
					mState=S_FINAL;
				}
				else if (mState==S_FINAL)
				{
				}	
				else  throw new IllegalArgumentException();
				break;
		}
	}
	
	private boolean waitForPending(long aTimeout)
	{	
		long t0 = System.currentTimeMillis();
		while (System.currentTimeMillis()-t0<aTimeout && mScheduler.hasPendingRequests())
		{	try
			{	Thread.sleep(100);
			} catch (InterruptedException e) {;}
		}
		return !mScheduler.hasPendingRequests();
	}
	
	private void hardStop(long aTimeout)
	{	mScheduler.stop(false);
		long t0 = System.currentTimeMillis();
		while (System.currentTimeMillis()-t0<aTimeout && !mScheduler.isStopped())
		{	try
			{	Thread.sleep(100);
			} catch (InterruptedException e) {;}
		}
		if (!mScheduler.isStopped())
		{	mScheduler.stop(true);
		}
	}
	
	public void interruptExternalRequests(Object aTransportInstance)
	{	if (mState==S_CLOSED)
		{	mRequestBlocker.interrupt(aTransportInstance);
			mRequestBlocker.releaseInterrupted();
		}
	}

	public void register(ISpaceWithIdentity aSpace) throws INetKernelException
	{	
		IIdentifier uri = aSpace.getIdentifier();
		IVersion v = aSpace.getVersion();
		if (getSpace(uri,v,v)==null)
		{	synchronized(mSpaces)
			{	mSpaces.put(uri,aSpace);
				mAreSpaceListenersDirty=true;
			}
		}
		else
		{	throw Utils.createFormattedException("EX_DUPLICATE_SPACE", "MSG_DUPLICATE_SPACE", mLogger, uri,v);
		}
	}

	public void deregister(ISpaceWithIdentity aSpace) 
	{	IIdentifier uri = aSpace.getIdentifier();
		boolean exists;
		synchronized(mSpaces)
		{	exists = mSpaces.remove(uri,aSpace);
		}
		if (exists)
		{	mAreSpaceListenersDirty=true;
		}
	}
	
	public void decommisionAll()
	{	
		if (mState==S_PAUSED)
		{
			synchronized(mSpaces)
			{	
				for (Iterator i=getSpaces(); i.hasNext(); )
				{
					ISpaceWithIdentity ctx = (ISpaceWithIdentity)i.next();
					mLogger.log(ILogger.INFO,this,"MSG_DEREGISTER",ctx.getIdentifier(),ctx.getVersion());
					try
					{
						ctx.onDecommissionSpace();
					}
					catch (Throwable e)
					{	mLogger.log(ILogger.WARNING,this,"ERR_DECOMMISSION",ctx.getIdentifier(),ctx.getVersion());
						e.printStackTrace();
						mLogger.logRaw(ILogger.WARNING,this,Utils.throwableToString(e));
					}
				}
				mSpaces.clear();
			}
		}
		else throw new IllegalStateException();
	}


	public Iterator<ISpaceWithIdentity> getSpaces()
	{	synchronized(mSpaces)
		{	List spaces = new ArrayList(mSpaces.size());
			for (Iterator i=mSpaces.valueIterator(); i.hasNext(); )
			{	spaces.add(i.next());
			}
			return spaces.iterator();
		}
	}


	public ISpaceWithIdentity getSpace(IIdentifier aIdentifier, IVersion aMin, IVersion aMax)
	{	ISpaceWithIdentity result = null;
		List spaces=mSpaces.get(aIdentifier);
		for (Iterator i=spaces.iterator(); i.hasNext(); )
		{	ISpaceWithIdentity space = (ISpaceWithIdentity)i.next();
			IVersion v = space.getVersion();
			if ( (aMin==null || aMin.isLessOrEqualThan(v)) && (aMax==null || aMax.isGreaterOrEqualThan(v)) )
			{	if (result==null || result.getVersion().isLessThan(v))
				{	result = space;
				}
			}
		}
		return result;
	}


	public void onAsyncRequest(IRequest aRequest)
	{	mScheduler.onAsyncRequest(aRequest);
	}
	
	public IResponse synchronousRequest(IRequest aRequest) throws INetKernelException
	{	return mScheduler.synchronousRequest(aRequest);
	}

	public void onAsyncExternalRequest(IRequest aRequest, Object aTransportInstance, ISpace aTransportSpace)
	{	try
		{	mRequestBlocker.check(aTransportInstance);
			mMonitor.externalRequest(aRequest,aTransportInstance,aTransportSpace);
			mScheduler.onAsyncRequest(aRequest);
		}
		catch (InterruptedException e)
		{	IResponse result = Utils.generateExceptionResponse(aRequest, this, "EX_NOT_ACCEPTING_REQUESTS");
			aRequest.getRequestor().onResponse( result );
		}
	}

	public IResponse synchronousExternalRequest(IRequest aRequest, Object aTransportInstance, ISpace aTransportSpace) throws INetKernelException
	{	try
		{	mRequestBlocker.check(aTransportInstance);
			mMonitor.externalRequest(aRequest,aTransportInstance,aTransportSpace);
			return mScheduler.synchronousRequest(aRequest);
		}
		catch (InterruptedException e)
		{	throw Utils.generateException(this,"EX_NOT_ACCEPTING_REQUESTS",null);
		}
	}	

	public ILogger getLogger()
	{	return mLogger;
	}
	
	public IConfiguration getConfiguration()
	{	return mConfiguration;
	}	
	
	public boolean inDebug()
	{	return mInDebug;
	}
	
	public void addSpaceListener(ISpaceListener aListener)
	{	synchronized(mSpaceListeners)
		{	mSpaceListeners.add(aListener);
		}
	}
	public void removeSpaceListener(ISpaceListener aListener)
	{	synchronized(mSpaceListeners)
		{	mSpaceListeners.remove(aListener);
		}
	}
	
	private void notifySpaceListeners()
	{	if (mAreSpaceListenersDirty)
		{	mAreSpaceListenersDirty=false;
			List listeners;
			synchronized(mSpaceListeners)
			{	listeners = new ArrayList(mSpaceListeners);
			}
			for (Iterator i=listeners.iterator(); i.hasNext(); )
			{	ISpaceListener listener = (ISpaceListener)i.next();
				try
				{	listener.spaceChanged();
				}
				catch( RuntimeException e)
				{	mLogger.log(ILogger.WARNING,this,"MSG_NOTIFY_SPACE",listener,Utils.throwableToString(e));
				}
			}
		}
	}

	public void addConfigurationListener(IConfigurationListener aListener)
	{	synchronized(mConfigurationListeners)
		{	mConfigurationListeners.add(aListener);
		}
		if (mConfiguration!=null)
		{	aListener.configurationChanged(mConfiguration);
		}
	}

	public void removeConfigurationListener(IConfigurationListener aListener)
	{	synchronized(mConfigurationListeners)
		{	mConfigurationListeners.remove(aListener);
		}
	}
		
	private void notifyConfigurationListeners()
	{	List listeners;
		synchronized(mConfigurationListeners)
		{	listeners = new ArrayList(mConfigurationListeners);
		}
		for (Iterator i=listeners.iterator(); i.hasNext(); )
		{	IConfigurationListener listener = (IConfigurationListener)i.next();
			try
			{	listener.configurationChanged(mConfiguration);
			}
			catch( RuntimeException e)
			{	mLogger.log(ILogger.WARNING,this,"MSG_NOTIFY_CONFIGURATION",listener,Utils.throwableToString(e));
			}
		}
	}
		
	public IKernelListener getKernelListener()
	{	return mMonitor;
	}
	
	public void setRepresentationCache(IRepresentationCache aCache)
	{	mRepresentationCache=aCache;
	}

	public IRepresentationCache getRepresentationCache()
	{	return mRepresentationCache;
	}
	
	public void setResolutionCache(IResolutionCache aCache)
	{	mResolutionCache=aCache;
	}
	
	public IResolutionCache getResolutionCache()
	{	return mResolutionCache;
	}
	
	public Scheduler getScheduler()
	{	return mScheduler;
	}

	public ITimer getTimer()
	{	return mTimer;
	}
}